class MissingError(Exception):
    """ Indicates something is missing """


class FailedAction(Exception):
    """ Indicates a specific action failed """


class InvalidAction(Exception):
    """ Indicates a specific action failed """
